import os
import json
import typing
import datetime
import pandas
import requests


class Core:

    '''
    Core functionality of :mod:`BharatFinTrack` module.
    '''

    def _excel_file_extension(
        self,
        file_path: str
    ) -> str:

        '''
        Returns the extension of an Excel file.

        Parameters
        ----------
        file_path : str
            Path of the Excel file.

        Returns
        -------
        str
            Extension of the Excel file.
        '''

        output = os.path.splitext(file_path)[-1]

        return output

    def string_to_date(
        self,
        date_string: str
    ) -> datetime.date:

        '''
        Converts a date string is in format 'DD-MMM-YYYY' to a `datetime.date` object.

        Parameters
        ----------
        date_string : str
            Date string in the format 'DD-MMM-YYYY'.

        Returns
        -------
        datetime.date
            A `datetime.date` object corresponding to the input date string.
        '''

        output = datetime.datetime.strptime(date_string, '%d-%b-%Y').date()

        return output

    @property
    def default_http_headers(
        self,
    ) -> dict[str, str]:

        '''
        Returns the default http headers to be used for the web requests.
        '''

        output = {
            'Content-Type': 'application/json; charset=UTF-8',
            'Accept': 'application/json, text/javascript, */*; q=0.01',
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/128.0.0.0 Safari/537.36',
            'Origin': 'https://www.niftyindices.com',
            'Accept-Encoding': 'gzip, deflate, br, zstd',
            'Connection': 'keep-alive',
            'X-Requested-With': 'XMLHttpRequest'
        }

        return output

    @property
    def url_nse_index_tri_data(
        self,
    ) -> str:

        '''
        Returns the url to access TRI (Total Return Index) data of NSE equity indices.
        '''

        output = 'https://www.niftyindices.com/Backpage.aspx/getTotalReturnIndexString'

        return output

    def _download_nse_tri(
        self,
        index_api: str,
        start_date: str,
        end_date: str,
        index: str,
        http_headers: typing.Optional[dict[str, str]] = None
    ) -> pandas.DataFrame:

        '''
        Helper method for the :meth:`NSETRI.download_historical_daily_data`.
        '''

        # payloads
        parameters = {
            'name': index_api,
            'startDate': start_date,
            'endDate': end_date,
            'indexName': index
        }
        payload = json.dumps(
            {
                'cinfo': json.dumps(parameters)
            }
        )

        # web request headers
        headers = self.default_http_headers if http_headers is None else http_headers

        # sent web requets
        response = requests.post(
            url=self.url_nse_index_tri_data,
            headers=headers,
            data=payload,
            timeout=15
        )
        response_data = response.json()
        records = json.loads(response_data['d'])
        df = pandas.DataFrame.from_records(records)
        df = df.iloc[:, -2:][::-1].reset_index(drop=True)
        df['Date'] = df['Date'].apply(
            lambda x: datetime.datetime.strptime(x, '%d %b %Y').date()
        )
        df = df.rename(columns={'TotalReturnsIndex': 'Close'})
        df['Close'] = df['Close'].astype(float)

        return df



import typing
import datetime
import dateutil.relativedelta
import pandas
import matplotlib
import warnings
from .nse_product import NSEProduct
from .core import Core


class NSETRI:

    '''
    Download and analyze NSE TRI (Total Return Index) data,
    including both price index and dividend reinvestment.
    '''

    @property
    def _index_api(
        self
    ) -> dict[str, str]:

        '''
        Returns a dictionary containing equity indices as keys
        and corresponding API names as values.
        '''

        df = NSEProduct()._dataframe_equity_index
        output = dict(
            zip(df['Index Name'], df['API TRI'])
        )

        return output

    @property
    def non_open_source_indices(
        self
    ) -> list[str]:

        '''
        Returns a list of equity indices that are not open-source.
        '''

        df = NSEProduct()._dataframe_equity_index
        df = df[df['API TRI'] == 'NON OPEN SOURCE']
        output = list(df['Index Name'].sort_values())

        return output

    def is_index_open_source(
        self,
        index: str,
    ) -> bool:

        '''
        Check whether the index data is open-source.

        Parameters
        ----------
        index : str
            Name of the index.

        Returns
        -------
        bool
            True if the index data is open-source, False otherwise.
        '''

        if NSEProduct().is_index_exist(index) is True:
            pass
        else:
            raise Exception(f'"{index}" index does not exist.')

        output = index not in self.non_open_source_indices

        return output

    def download_historical_daily_data(
        self,
        index: str,
        start_date: typing.Optional[str] = None,
        end_date: typing.Optional[str] = None,
        http_headers: typing.Optional[dict[str, str]] = None,
        excel_file: typing.Optional[str] = None
    ) -> pandas.DataFrame:

        '''
        Downloads historical daily closing values for the specified index
        between the given start and end dates, both inclusive.

        Parameters
        ----------
        index : str
            Name of the index.

        start_date : str, optional
            Start date in the format 'DD-MMM-YYYY'.
            Defaults to the index's base date if None is provided.

        end_date : str, optional
            End date in the format 'DD-MMM-YYYY'.
            Defaults to the current date if None is provided.

        http_headers : dict, optional
            HTTP headers for the web request. If not provided, defaults to
            :attr:`BharatFinTrack.core.Core.default_http_headers`.

        excel_file : str, optional
            Path to an Excel file to save the DataFrame.

        Returns
        -------
        DataFrame
            A DataFrame containing the daily closing values for the index between the specified dates.
        '''

        # check index name
        if self.is_index_open_source(index) is True:
            index_api = self._index_api.get(index, index)
        else:
            raise Exception(f'"{index}" index data is not available as open-source.')

        # check start date
        if start_date is not None:
            pass
        else:
            start_date = NSEProduct().get_equity_index_base_date(index)
        date_s = Core().string_to_date(start_date)

        # check end date
        if end_date is not None:
            pass
        else:
            end_date = datetime.date.today().strftime('%d-%b-%Y')
        date_e = Core().string_to_date(end_date)

        # check end date is greater than start date
        difference_days = (date_e - date_s).days
        if difference_days >= 0:
            pass
        else:
            raise Exception(f'Start date {start_date} cannot be later than end date {end_date}.')

        # check the Excel file extension first
        excel_ext = Core()._excel_file_extension(excel_file) if excel_file is not None else None
        if excel_ext is None or excel_ext == '.xlsx':
            pass
        else:
            raise Exception(f'Input file extension "{excel_ext}" does not match the required ".xlsx".')

        # downloaded DataFrame
        df = Core()._download_nse_tri(
            index_api=index_api,
            start_date=start_date,
            end_date=end_date,
            index=index,
            http_headers=http_headers
        )

        # saving the DataFrame
        if excel_ext is None:
            pass
        else:
            with pandas.ExcelWriter(excel_file, engine='xlsxwriter') as excel_writer:
                df.to_excel(excel_writer, index=False)
                worksheet = excel_writer.sheets['Sheet1']
                worksheet.set_column(0, 1, 12)

        return df

    def update_historical_daily_data(
        self,
        index: str,
        excel_file: str,
        http_headers: typing.Optional[dict[str, str]] = None
    ) -> pandas.DataFrame:

        '''
        Updates historical daily closing values from the last date in the input Excel file
        to the present and saves the aggregated data to the same file.

        Parameters
        ----------
        index : str
            Name of the index.

        excel_file : str
            Path to the Excel file containing existing historical data.

        http_headers : dict, optional
            HTTP headers for the web request. If not provided, defaults to
            :attr:`BharatFinTrack.core.Core.default_http_headers`.

        Returns
        -------
        DataFrame
            A DataFrame with updated closing values from the last recorded date to the present.
        '''

        # read the input Excel file
        df = pandas.read_excel(excel_file)
        df['Date'] = df['Date'].apply(
            lambda x: x.date()
        )

        # addition of downloaded DataFrame
        add_df = self.download_historical_daily_data(
            index=index,
            start_date=df.iloc[-1, 0].strftime('%d-%b-%Y'),
            end_date=datetime.date.today().strftime('%d-%b-%Y'),
            http_headers=http_headers
        )

        # updating the DataFrame
        update_df = pandas.concat([df, add_df]) if isinstance(add_df, pandas.DataFrame) else df
        update_df = update_df.drop_duplicates().reset_index(drop=True)

        # saving the DataFrame
        with pandas.ExcelWriter(excel_file, engine='xlsxwriter') as excel_writer:
            update_df.to_excel(excel_writer, index=False)
            worksheet = excel_writer.sheets['Sheet1']
            worksheet.set_column(0, 1, 12)

        return add_df

    def download_daily_summary_equity_closing(
        self,
        excel_file: str,
        http_headers: typing.Optional[dict[str, str]] = None,
        test_mode: bool = False
    ) -> pandas.DataFrame:

        '''
        Returns updated TRI closing values for all NSE indices.

        Parameters
        ----------
        excel_file : str, optional
            Path to an Excel file to save the DataFrame.

        http_headers : dict, optional
            HTTP headers for the web request. Defaults to
            :attr:`BharatFinTrack.core.Core.default_http_headers` if not provided.

        test_mode : bool, optional
            If True, the function will use a mocked DataFrame for testing purposes
            instead of the actual data. This parameter is intended for developers
            for testing purposes only and is not recommended for use by end-users.

        Returns
        -------
        DataFrame
            A DataFrame containing updated TRI closing values for all NSE indices.
        '''

        # processing base DataFrame
        base_df = NSEProduct()._dataframe_equity_index
        base_df = base_df.groupby(level='Category').head(1) if test_mode is True else base_df
        base_df = base_df.reset_index()
        base_df = base_df.drop(columns=['ID', 'API TRI'])
        base_df['Base Date'] = base_df['Base Date'].apply(lambda x: x.date())

        # check the Excel file extension first
        excel_ext = Core()._excel_file_extension(excel_file)
        if excel_ext == '.xlsx':
            pass
        else:
            raise Exception(f'Input file extension "{excel_ext}" does not match the required ".xlsx".')

        # downloading data
        today = datetime.date.today()
        week_ago = today - datetime.timedelta(days=7)
        end_date = today.strftime('%d-%b-%Y')
        start_date = week_ago.strftime('%d-%b-%Y')
        for base_index in base_df.index:
            try:
                index_df = self.download_historical_daily_data(
                    index=base_df.loc[base_index, 'Index Name'],
                    start_date=start_date,
                    end_date=end_date
                )
                base_df.loc[base_index, 'Close Date'] = index_df.iloc[-1, 0]
                base_df.loc[base_index, 'Close Value'] = index_df.iloc[-1, -1]
            except Exception:
                base_df.loc[base_index, 'Close Date'] = end_date
                base_df.loc[base_index, 'Close Value'] = -1000

        # removing error rows from the DataFrame
        base_df = base_df[base_df['Close Value'] != -1000].reset_index(drop=True)

        # saving the DataFrame
        with pandas.ExcelWriter(excel_file, engine='xlsxwriter') as excel_writer:
            base_df.to_excel(excel_writer, index=False)
            worksheet = excel_writer.sheets['Sheet1']
            # format columns
            for col_num, df_col in enumerate(base_df.columns):
                if df_col == 'Index Name':
                    worksheet.set_column(col_num, col_num, 60)
                else:
                    worksheet.set_column(col_num, col_num, 15)

        return base_df

    def download_equity_indices_updated_value(
        self,
        excel_file: str,
        http_headers: typing.Optional[dict[str, str]] = None,
        test_mode: bool = False
    ) -> pandas.DataFrame:

        '''
        Returns updated TRI values for all NSE indices.

        Parameters
        ----------
        excel_file : str, optional
            Path to an Excel file to save the DataFrame.

        http_headers : dict, optional
            HTTP headers for the web request. Defaults to
            :attr:`BharatFinTrack.core.Core.default_http_headers` if not provided.

        test_mode : bool, optional
            If True, the function will use a mocked DataFrame for testing purposes
            instead of the actual data. This parameter is intended for developers
            for testing purposes only and is not recommended for use by end-users.

        Returns
        -------
        DataFrame
            A DataFrame containing updated TRI values for all NSE indices.

        Warnings
        --------
        The method :meth:`BharatFinTrack.NSETRI.download_equity_indices_updated_value` is deprecated.
        Use the :meth:`BharatFinTrack.NSETRI.download_daily_summary_equity_closing` instead.
        '''

        warnings.warn(
            'download_equity_indices_updated_value(excel_file) is deprecated. Use download_daily_summary_equity_closing(excel_file) instead.',
            DeprecationWarning,
            stacklevel=2
        )

        output = self.download_daily_summary_equity_closing(
            excel_file=excel_file,
            http_headers=http_headers,
            test_mode=test_mode
        )

        return output

    def sort_equity_value_from_launch(
        self,
        input_excel: str,
        output_excel: str,
    ) -> pandas.DataFrame:

        '''
        Returns equity indices sorted in descending order by TRI values since launch.

        Parameters
        ----------
        inout_excel : str
            Path to the input Excel file.

        output_excel : str
            Path to an output Excel file to save the output DataFrame.

        Returns
        -------
        DataFrame
            A DataFrame sorted in descending order by TRI values since launch.
        '''

        # sorting DataFrame by TRI values
        df = pandas.read_excel(input_excel)
        df = df.drop(columns=['Category'])
        df = df.sort_values(
            by=['Close Value'],
            ascending=[False]
        )
        df = df.reset_index(drop=True)
        for col_df in df.columns:
            if 'Date' in col_df:
                df[col_df] = df[col_df].apply(lambda x: x.date())
            else:
                pass

        # saving the DataFrame
        excel_ext = Core()._excel_file_extension(output_excel)
        if excel_ext != '.xlsx':
            raise Exception(
                f'Input file extension "{excel_ext}" does not match the required ".xlsx".'
            )
        else:
            with pandas.ExcelWriter(output_excel, engine='xlsxwriter') as excel_writer:
                df.to_excel(excel_writer, index=False)
                worksheet = excel_writer.sheets['Sheet1']
                # format columns
                for col_num, col_df in enumerate(df.columns):
                    if col_df == 'Index Name':
                        worksheet.set_column(col_num, col_num, 60)
                    else:
                        worksheet.set_column(col_num, col_num, 15)

        return df

    def sort_equity_cagr_from_launch(
        self,
        input_excel: str,
        output_excel: str,
    ) -> pandas.DataFrame:

        '''
        Returns equity indices sorted in descending order by CAGR (%) since launch.

        Parameters
        ----------
        inout_excel : str
            Path to the input Excel file.

        output_excel : str
            Path to an output Excel file to save the output DataFrame.

        Returns
        -------
        DataFrame
            A DataFrame sorted in descending order by CAGR (%) values since launch.
        '''

        # DataFrame processing
        df = pandas.read_excel(input_excel)
        df = df.drop(columns=['Category'])
        for col_df in df.columns:
            if 'Date' in col_df:
                df[col_df] = df[col_df].apply(lambda x: x.date())
            else:
                pass
        df['Close/Base'] = df['Close Value'] / df['Base Value']
        df['Years'] = list(
            map(
                lambda x, y: dateutil.relativedelta.relativedelta(x, y).years, df['Close Date'], df['Base Date']
            )
        )
        df['Days'] = list(
            map(
                lambda x, y, z: (x - y.replace(year=y.year + z)).days, df['Close Date'], df['Base Date'], df['Years']
            )
        )
        total_years = df['Years'] + (df['Days'] / 365)
        df['CAGR(%)'] = 100 * (pow(df['Close Value'] / df['Base Value'], 1 / total_years) - 1)

        # sorting DataFrame by CAGR (%) values
        df = df.sort_values(
            by=['CAGR(%)', 'Years', 'Days'],
            ascending=[False, False, False]
        )
        df = df.reset_index(drop=True)

        # saving the DataFrame
        excel_ext = Core()._excel_file_extension(output_excel)
        if excel_ext != '.xlsx':
            raise Exception(
                f'Input file extension "{excel_ext}" does not match the required ".xlsx".'
            )
        else:
            with pandas.ExcelWriter(output_excel, engine='xlsxwriter') as excel_writer:
                df.to_excel(excel_writer, index=False)
                workbook = excel_writer.book
                worksheet = excel_writer.sheets['Sheet1']
                # format columns
                for col_num, col_df in enumerate(df.columns):
                    if col_df == 'Index Name':
                        worksheet.set_column(col_num, col_num, 60)
                    elif col_df == 'Close Value':
                        worksheet.set_column(
                            col_num, col_num, 15,
                            workbook.add_format({'num_format': '#,##0'})
                        )
                    elif col_df == 'Close/Base':
                        worksheet.set_column(
                            col_num, col_num, 15,
                            workbook.add_format({'num_format': '#,##0.0'})
                        )
                    elif col_df == 'CAGR(%)':
                        worksheet.set_column(
                            col_num, col_num, 15,
                            workbook.add_format({'num_format': '#,##0.00'})
                        )
                    else:
                        worksheet.set_column(col_num, col_num, 15)

        return df

    def category_sort_equity_cagr_from_launch(
        self,
        input_excel: str,
        output_excel: str,
    ) -> pandas.DataFrame:

        '''
        Returns equity indices sorted in descending order by CAGR (%) since launch
        within each index category.

        Parameters
        ----------
        inout_excel : str
            Path to the input Excel file.

        output_excel : str
            Path to an output Excel file to save the output DataFrame.

        Returns
        -------
        DataFrame
            A multi-index DataFrame sorted in descending order by CAGR (%) values since launch
            within each index category.
        '''

        # DataFrame processing
        df = pandas.read_excel(input_excel)
        for col_df in df.columns:
            if 'Date' in col_df:
                df[col_df] = df[col_df].apply(lambda x: x.date())
            else:
                pass
        df['Close/Base'] = df['Close Value'] / df['Base Value']
        df['Years'] = list(
            map(
                lambda x, y: dateutil.relativedelta.relativedelta(x, y).years, df['Close Date'], df['Base Date']
            )
        )
        df['Days'] = list(
            map(
                lambda x, y, z: (x - y.replace(year=y.year + z)).days, df['Close Date'], df['Base Date'], df['Years']
            )
        )
        total_years = df['Years'] + (df['Days'] / 365)
        df['CAGR(%)'] = 100 * (pow(df['Close Value'] / df['Base Value'], 1 / total_years) - 1)

        # Convert 'Category' column to categorical data types with a defined order
        categories = list(df['Category'].unique())
        df['Category'] = pandas.Categorical(
            df['Category'],
            categories=categories,
            ordered=True
        )

        # Sorting Dataframe
        df = df.sort_values(
            by=['Category', 'CAGR(%)', 'Years', 'Days'],
            ascending=[True, False, False, False]
        )
        dataframes = []
        for category in categories:
            category_df = df[df['Category'] == category]
            category_df = category_df.drop(columns=['Category']).reset_index(drop=True)
            dataframes.append(category_df)
        output = pandas.concat(
            dataframes,
            keys=[word.upper() for word in categories],
            names=['Category', 'ID']
        )

        # saving the DataFrame
        excel_ext = Core()._excel_file_extension(output_excel)
        if excel_ext != '.xlsx':
            raise Exception(
                f'Input file extension "{excel_ext}" does not match the required ".xlsx".'
            )
        else:
            with pandas.ExcelWriter(output_excel, engine='xlsxwriter') as excel_writer:
                output.to_excel(excel_writer, index=True)
                workbook = excel_writer.book
                worksheet = excel_writer.sheets['Sheet1']
                # number of columns for DataFrame indices
                index_cols = len(output.index.names)
                # format columns
                worksheet.set_column(0, index_cols - 1, 15)
                for col_num, col_df in enumerate(output.columns):
                    if col_df == 'Index Name':
                        worksheet.set_column(index_cols + col_num, index_cols + col_num, 60)
                    elif col_df == 'Close Value':
                        worksheet.set_column(
                            index_cols + col_num, index_cols + col_num, 15,
                            workbook.add_format({'num_format': '#,##0'})
                        )
                    elif col_df == 'Close/Base':
                        worksheet.set_column(
                            index_cols + col_num, index_cols + col_num, 15,
                            workbook.add_format({'num_format': '#,##0.0'})
                        )
                    elif col_df == 'CAGR(%)':
                        worksheet.set_column(
                            index_cols + col_num, index_cols + col_num, 15,
                            workbook.add_format({'num_format': '#,##0.00'})
                        )
                    else:
                        worksheet.set_column(index_cols + col_num, index_cols + col_num, 15)
                # Dataframe colors
                get_colormap = matplotlib.colormaps.get_cmap('Pastel2')
                colors = [
                    get_colormap(count / len(dataframes)) for count in range(len(dataframes))
                ]
                hex_colors = [
                    '{:02X}{:02X}{:02X}'.format(*[int(num * 255) for num in color]) for color in colors
                ]
                # coloring of DataFrames
                start_col = index_cols - 1
                end_col = index_cols + len(output.columns) - 1
                start_row = 1
                for df, color in zip(dataframes, hex_colors):
                    color_format = workbook.add_format({'bg_color': color})
                    end_row = start_row + len(df) - 1
                    worksheet.conditional_format(
                        start_row, start_col, end_row, end_col,
                        {'type': 'no_blanks', 'format': color_format}
                    )
                    start_row = end_row + 1

        return output



import pytest
import BharatFinTrack
import os
import tempfile
import datetime
import pandas


@pytest.fixture(scope='class')
def nse_product():

    yield BharatFinTrack.NSEProduct()


@pytest.fixture(scope='class')
def nse_index():

    yield BharatFinTrack.NSEIndex()


@pytest.fixture(scope='class')
def nse_tri():

    yield BharatFinTrack.NSETRI()


@pytest.fixture(scope='class')
def core():

    yield BharatFinTrack.core.Core()


@pytest.fixture
def message():

    output = {
        'error_category': 'Input category "region" does not exist.',
        'error_date1': "time data '16-Sep-202' does not match format '%d-%b-%Y'",
        'error_date2': "time data '20-Se-2024' does not match format '%d-%b-%Y'",
        'error_date3': 'Start date 27-Sep-2024 cannot be later than end date 26-Sep-2024.',
        'error_excel': 'Input file extension ".xl" does not match the required ".xlsx".',
        'error_folder': 'The folder path does not exist.',
        'error_index1': '"INVALID" index does not exist.',
        'error_index2': '"NIFTY50 USD" index data is not available as open-source.'

    }

    return output


def test_save_dataframes_equity_indices(
    nse_product,
    message
):

    # error test
    with pytest.raises(Exception) as exc_info:
        nse_product.save_dataframe_equity_index_parameters(
            excel_file=r"C:\Users\Username\Folder\out.xl"
        )
    assert exc_info.value.args[0] == message['error_excel']

    # pass test
    with tempfile.TemporaryDirectory() as tmp_dir:
        excel_file = os.path.join(tmp_dir, 'equity.xlsx')
        df = nse_product.save_dataframe_equity_index_parameters(
            excel_file=excel_file
        )
        assert len(df.index.names) == 2


def test_get_equity_indices_by_category(
    nse_product,
    message
):

    # pass test
    assert 'NIFTY 500' in nse_product.get_equity_indices_by_category('broad')
    assert 'NIFTY IT' in nse_product.get_equity_indices_by_category('sector')
    assert 'NIFTY HOUSING' in nse_product.get_equity_indices_by_category('thematic')
    assert 'NIFTY ALPHA 50' in nse_product.get_equity_indices_by_category('strategy')
    assert 'NIFTY50 USD' in nse_product.get_equity_indices_by_category('variant')

    # error test
    with pytest.raises(Exception) as exc_info:
        nse_product.get_equity_indices_by_category('region')
    assert exc_info.value.args[0] == message['error_category']


def test_is_index_exist(
    nse_product
):

    assert nse_product.is_index_exist('NIFTY 100') is True
    assert nse_product.is_index_exist('INVALID') is False


def test_get_equity_index_base_date(
    nse_product,
    message
):

    # pass test
    assert nse_product.get_equity_index_base_date('NIFTY100 EQUAL WEIGHT') == '01-Jan-2003'
    assert nse_product.get_equity_index_base_date('NIFTY INDIA DEFENCE') == '02-Apr-2018'

    # error test
    with pytest.raises(Exception) as exc_info:
        nse_product.get_equity_index_base_date('INVALID')
    assert exc_info.value.args[0] == message['error_index1']


def test_get_equity_index_base_value(
    nse_product,
    message
):

    # pass test
    assert nse_product.get_equity_index_base_value('NIFTY MIDCAP LIQUID 15') == 1500.0
    assert nse_product.get_equity_index_base_value('NIFTY IT') == 100.0

    # error test
    with pytest.raises(Exception) as exc_info:
        nse_product.get_equity_index_base_value('INVALID')
    assert exc_info.value.args[0] == message['error_index1']


def test_is_index_data_open_source(
    nse_tri,
    message
):

    # pass test
    assert nse_tri.is_index_open_source('NIFTY 50') is True
    assert nse_tri.is_index_open_source('NIFTY50 USD') is False

    # error test
    with pytest.raises(Exception) as exc_info:
        nse_tri.is_index_open_source('INVALID')
    assert exc_info.value.args[0] == message['error_index1']


def test_download_historical_daily_data(
    nse_tri,
    message
):

    # error test for non open-source index
    with pytest.raises(Exception) as exc_info:
        nse_tri.download_historical_daily_data(
            index='NIFTY50 USD',
            start_date='27-Sep-2024',
            end_date='27-Sep-2024'
        )
    assert exc_info.value.args[0] == message['error_index2']

    # error test for invalid start date input
    with pytest.raises(Exception) as exc_info:
        nse_tri.download_historical_daily_data(
            index='NIFTY 50',
            start_date='16-Sep-202',
            end_date='26-Sep-2024'
        )
    assert exc_info.value.args[0] == message['error_date1']

    # error test for invalid end date input
    with pytest.raises(Exception) as exc_info:
        nse_tri.download_historical_daily_data(
            index='NIFTY 50',
            start_date='16-Sep-2024',
            end_date='20-Se-2024'
        )
    assert exc_info.value.args[0] == message['error_date2']

    # error test for strat date later than end date
    with pytest.raises(Exception) as exc_info:
        nse_tri.download_historical_daily_data(
            index='NIFTY 50',
            start_date='27-Sep-2024',
            end_date='26-Sep-2024'
        )
    assert exc_info.value.args[0] == message['error_date3']

    # error test for invalid Excel file input
    with pytest.raises(Exception) as exc_info:
        nse_tri.download_historical_daily_data(
            index='NIFTY 50',
            start_date='23-Sep-2024',
            end_date='27-Sep-2024',
            excel_file='NIFTY50_tri.xl'
        )
    assert exc_info.value.args[0] == message['error_excel']

    # pass test for start date being None
    df = nse_tri.download_historical_daily_data(
        index='NIFTY INDIA DEFENCE',
        start_date=None,
        end_date='10-Apr-2018'
    )
    assert float(df.iloc[0, -1]) == 1000.00

    # pass test for end date being None
    start_date = (datetime.date.today() - datetime.timedelta(days=7)).strftime('%d-%b-%Y')
    df = nse_tri.download_historical_daily_data(
        index='NIFTY 50',
        start_date=start_date,
        end_date=None
    )
    assert len(df) > 0


@pytest.mark.parametrize(
    'index, expected_value',
    [
        ('NIFTY 50', 38861.64),
        ('NIFTY MIDCAP150 MOMENTUM 50', 82438.16)
    ]
)
def test_index_download_historical_daily_data(
    nse_tri,
    index,
    expected_value
):

    df = nse_tri.download_historical_daily_data(
        index=index,
        start_date='27-Sep-2024',
        end_date='27-Sep-2024'
    )
    assert float(df.iloc[-1, -1]) == expected_value


def test_equity_index_price_download_updated_value(
    nse_index,
    message,
    capsys
):

    # error test for folder path
    with tempfile.TemporaryDirectory() as tmp_dir:
        pass
    with pytest.raises(Exception) as exc_info:
        nse_index.download_daily_summary_report(tmp_dir)
    assert exc_info.value.args[0] == message['error_folder']
    # pass test for capturing print statement
    nse_index.equity_cagr_from_launch(
        untracked_indices=True
    )
    capture_print = capsys.readouterr()
    assert 'List of untracked download indices' in capture_print.out
    assert 'List of untracked base indices' in capture_print.out

    # pass test for sorting of NSE equity indices by CAGR (%) value
    with tempfile.TemporaryDirectory() as tmp_dir:
        excel_file = os.path.join(tmp_dir, 'equity.xlsx')
        nse_index.sort_equity_cagr_from_launch(
            excel_file=excel_file
        )
        df = pandas.read_excel(excel_file)
        assert len(df.index.names) == 1
    # error test for invalid Excel file input
    with pytest.raises(Exception) as exc_info:
        nse_index.sort_equity_cagr_from_launch(
            excel_file='equily.xl'
        )
    assert exc_info.value.args[0] == message['error_excel']

    # pass test for categorical sorting NSE equity indices by CAGR (%) value
    with tempfile.TemporaryDirectory() as tmp_dir:
        excel_file = os.path.join(tmp_dir, 'equity.xlsx')
        nse_index.category_sort_equity_cagr_from_launch(
            excel_file=excel_file
        )
        df = pandas.read_excel(excel_file, index_col=[0, 1])
        assert df.shape[1] == 9
        assert len(df.index.get_level_values('Category').unique()) <= 5
    # error test for invalid Excel file input
    with pytest.raises(Exception) as exc_info:
        nse_index.category_sort_equity_cagr_from_launch(
            excel_file='equily.xl'
        )
    assert exc_info.value.args[0] == message['error_excel']


@pytest.mark.filterwarnings('ignore::DeprecationWarning')
def test_equity_index_tri_download_updated_value(
    nse_tri,
    message
):

    with tempfile.TemporaryDirectory() as tmp_dir:
        excel_file = os.path.join(tmp_dir, 'equity.xlsx')
        # pass test for downloading updated TRI values of NSE equity indices
        nse_tri.download_equity_indices_updated_value(
            excel_file=excel_file,
            test_mode=True
        )
        df = pandas.read_excel(excel_file)
        assert df.shape[1] == 6
        assert df.shape[0] <= 4
        # error test for invalid Excel file input
        with pytest.raises(Exception) as exc_info:
            nse_tri.download_equity_indices_updated_value(
                excel_file='output.xl'
            )
        assert exc_info.value.args[0] == message['error_excel']
        # pass test for sorting of NSE equity indices by TRI values
        output_excel = os.path.join(tmp_dir, 'sorted_tri_value.xlsx')
        nse_tri.sort_equity_value_from_launch(
            input_excel=excel_file,
            output_excel=output_excel
        )
        df = pandas.read_excel(output_excel)
        assert df.shape[1] == 5
        # error test for invalid Excel file input
        with pytest.raises(Exception) as exc_info:
            nse_tri.sort_equity_value_from_launch(
                input_excel=excel_file,
                output_excel='output.xl'
            )
        assert exc_info.value.args[0] == message['error_excel']
        # pass test for sorting of NSE equity indices by CAGR (%) value
        output_excel = os.path.join(tmp_dir, 'sorted_tri_cagr.xlsx')
        nse_tri.sort_equity_cagr_from_launch(
            input_excel=excel_file,
            output_excel=output_excel
        )
        df = pandas.read_excel(output_excel)
        assert df.shape[1] == 9
        # error test for invalid Excel file input
        with pytest.raises(Exception) as exc_info:
            nse_tri.sort_equity_cagr_from_launch(
                input_excel=excel_file,
                output_excel='output.xl'
            )
        assert exc_info.value.args[0] == message['error_excel']
        # pass test for categorical sorting NSE equity indices by CAGR (%) value
        output_excel = os.path.join(tmp_dir, 'categorical_sorted_tri_cagr.xlsx')
        nse_tri.category_sort_equity_cagr_from_launch(
            input_excel=excel_file,
            output_excel=output_excel
        )
        df = pandas.read_excel(output_excel, index_col=[0, 1])
        assert len(df.index.get_level_values('Category').unique()) <= 4
        # error test for invalid Excel file input
        with pytest.raises(Exception) as exc_info:
            nse_tri.category_sort_equity_cagr_from_launch(
                input_excel=excel_file,
                output_excel='output.xl'
            )
        assert exc_info.value.args[0] == message['error_excel']


def test_update_historical_daily_data(
    nse_tri
):

    with tempfile.TemporaryDirectory() as tmp_dir:
        excel_file = os.path.join(tmp_dir, 'equity.xlsx')
        today = datetime.date.today()
        day1_ago = today - datetime.timedelta(days=30)
        day2_ago = today - datetime.timedelta(days=15)
        # pass test for downloading daily TRI values of NSE a equity index
        nse_tri.download_historical_daily_data(
            index='NIFTY INDIA DEFENCE',
            start_date=day1_ago.strftime('%d-%b-%Y'),
            end_date=day2_ago.strftime('%d-%b-%Y'),
            excel_file=excel_file
        )
        len_df1 = len(pandas.read_excel(excel_file))
        # pass test for updating daily TRI values for the NSE equity index
        nse_tri.update_historical_daily_data(
            index='NIFTY INDIA DEFENCE',
            excel_file=excel_file
        )
        len_df2 = len(pandas.read_excel(excel_file))
        assert len_df2 > len_df1
